from web_framework_v2.http import HttpMethod


class HttpRequest:
    def __init__(self, http_method: HttpMethod, url: str, version: str, headers: dict, query_parameters: dict, body: bytes, path_variables=None):
        self.method = http_method
        self.url: str = url
        self.http_version: str = version
        self.headers = headers if headers is not None else {}
        self.query_parameters = query_parameters if query_parameters is not None else {}
        self.body = body
        self.path_variables = path_variables

    def get_header(self, name: str):
        return self.headers.get(name, None)

    def get_query_parameter(self, name: str):
        return self.query_parameters.get(name, None)

    def clone(self):
        return HttpRequest(
            self.method,
            self.url,
            self.http_version,
            dict(self.headers) if self.headers is not None else None,
            dict(self.query_parameters) if self.query_parameters is not None else None,
            self.body,
            self.path_variables
        )

    def __str__(self):
        return f"HttpRequest(method: {self.method}, url: {self.url}, http_version: {self.http_version}, headers: {self.headers}, query_parameters: {self.query_parameters}"



import time
from abc import ABC
from datetime import datetime, timezone
from typing import Tuple

import jwt
from web_framework_v2.decorator import Decorator


class JwtSecurity(Decorator, ABC):
    _SECRET = "secret_key_temporary"

    def __init__(self, on_fail=lambda request, response: None):
        """
        Base class of JWT security
        :param on_fail: called when security check fails.
        """
        self.on_fail = on_fail

    @staticmethod
    def secret():
        return JwtSecurity._SECRET

    @staticmethod
    def set_secret(secret):
        JwtSecurity._SECRET = secret

    @staticmethod
    def decode_token(token):
        try:
            return jwt.decode(token, JwtSecurity.secret(), algorithms=["HS256"])
        except:
            return None

    @staticmethod
    def decode_request(request):
        return JwtSecurity.decode_token(request.headers["Authorization"][8:])

    @staticmethod
    def create_token(data, expiration_seconds: int):
        data["exp"] = datetime.now(tz=timezone.utc).timestamp() + expiration_seconds
        return jwt.encode(data, JwtSecurity.secret())


class JwtTokenFactory(JwtSecurity, ABC):
    """
    Authenticates and creates a new token using request and request body
    """

    def __init__(self, on_fail=lambda request, response: None, expiration_time=60 * 30):
        super().__init__(on_fail=on_fail)
        self.expiration_time = expiration_time

    def should_execute_endpoint(self, request, request_body) -> Tuple[bool, object, object]:
        authentication_result, authentication_data, builder_data = self.authenticate(request, request_body)
        if not authentication_result:
            return False, None, authentication_data

        return True, JwtSecurity.create_token(self.token_data_builder(request, request_body, builder_data), self.expiration_time), authentication_data

    def token_data_builder(self, request, request_body, data):
        raise NotImplementedError(f"Must implement token data builder for {type(self)}!")

    def authenticate(self, request, request_body) -> (bool, object):
        raise NotImplementedError(f"Must implement request authentication for {type(self)}!")


class JwtTokenAuth(JwtSecurity, ABC):
    """
    Allows access and authenticates using a pre-existing token.
    """

    def should_execute_endpoint(self, request, request_body) -> Tuple[bool, object, object]:
        decoded_token = JwtSecurity.decode_request(request)
        authentication_result, authentication_data = self.authenticate(request, request_body, decoded_token)

        return authentication_result, self.decoded_token_transformer(request, request_body, decoded_token), authentication_data

    def authenticate(self, request, request_body, decoded_token) -> (bool, object):
        raise NotImplementedError(f"Must implement authentication for {type(self)}!")

    def decoded_token_transformer(self, request, request_body, decoded_token) -> (bool, object):
        raise NotImplementedError(f"Must implement decoded_token_transformer for {type(self)}!")


from enum import Enum


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    COPY = "COPY"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    LINK = "LINK"
    UNLINK = "UNLINK"
    PURGE = "PURGE"
    LOCK = "LOCK"
    UNLOCK = "UNLOCK"
    PROPFIND = "PROPFIND"
    VIEW = "VIEW"


import logging
import os
import time
import traceback

import web_framework_v2.route.endpoint as endpoint
from web_framework_v2.http import HttpStatus, ContentType

logger = logging.getLogger(__name__)


class HttpResponse:
    def __init__(self, content_type, http_version: str, status, html: bytes):
        self.content_type = str(content_type)
        self.http_version = http_version
        self.status = status
        self.html = html if html is not None else b""
        self._data: bytes = bytes()

    def data(self):
        self._data = self._build_response()
        return self._data

    def _build_response(self, receive_time=None):
        response = f"{self.http_version.strip()} {self.status.value} {self.status.name}\r\n".encode()
        response += f"{self.content_type}\r\n".encode()
        response += self.build_header("Content-Length", len(self.html))
        if receive_time is not None:
            response += self.build_header("Server-Timing", str(time.time() - receive_time))
        response += '\r\n'.encode()
        response += self.html
        return response

    @staticmethod
    def build_header(header, value):
        return f"{header}: {value}\r\n".encode()

    @staticmethod
    def build_from_route(request, route: endpoint.Endpoint, path_variables: dict):
        try:
            response = HttpResponse.build_empty_status_response(request, HttpStatus.OK, b"")
            logger.debug(f"Executing route {route.route()} with url {request.url}")
            res = route.execute(request.clone(), response, path_variables)
            logger.debug(f"Successfully executed route {route.route()} with url {request.url}\nResult: {res}")
            return HttpResponse(route.content_type(), response.http_version, response.status, str(res).encode() if type(res) is not bytes else res)
        except Exception as e:
            logger.exception(e)
            return HttpResponse(ContentType.text, request.http_version, HttpStatus.INTERNAL_SERVER_ERROR, traceback.format_exc().encode())

    @staticmethod
    def build_empty_status_response(request, status: HttpStatus, additional_info: bytes):
        if additional_info is None:
            additional_info = ""
        return HttpResponse(ContentType.html, request.http_version, status, additional_info)

    @staticmethod
    def build_from_file(request, path: str):
        try:
            if not os.path.isfile(path):
                logger.debug(f"Failed to find static file at {path}")
                return HttpResponse(ContentType.text, request.http_version, HttpStatus.NOT_FOUND, bytes())
            else:
                extension = os.path.splitext(path)[1][1::]
                content_type = ContentType[extension]
                html: bytes
                logger.debug(f"Building response using static file at {path}")
                with open(path, 'rb') as file:
                    html = file.read() + b"\r\n\r\n"
                return HttpResponse(content_type, request.http_version, HttpStatus.OK, html)
        except Exception as e:
            logger.exception(e)
            return HttpResponse(ContentType.text, request.http_version, HttpStatus.INTERNAL_SERVER_ERROR, traceback.format_exc().encode())



import logging
import socket
import threading

from web_framework_v2.parser import RequestParser

logger = logging.getLogger(__name__)


class HttpClient:
    def __init__(self, client_socket: socket.socket, address, response_builder):
        self.socket = client_socket
        self.address = address
        self.response_builder = response_builder
        self.response_receive_time = 0
        self.is_closed = False
        self.response_handler_thread = threading.Thread(target=self.request_handler)
        self.byte_fetch_amount = 2048

    def start(self):
        self.response_handler_thread.start()

    def close(self):
        self.socket.shutdown(socket.SHUT_WR)
        logger.debug(f"Closing client socket and exiting client thread {threading.current_thread()}")
        exit()

    def send(self, data: bytes):
        self.socket.sendall(data)

    def request_handler(self):
        while not self.is_closed:
            data: bytes = self.socket.recv(self.byte_fetch_amount)
            curr_data_buffer = data
            last_bytes = False
            while len(curr_data_buffer) == self.byte_fetch_amount and not last_bytes:
                last_bytes = len(self.socket.recv(self.byte_fetch_amount + 1,
                                                  socket.MSG_PEEK)) <= self.byte_fetch_amount  # prevent blocking if cycle equals fetch amount
                curr_data_buffer = self.socket.recv(self.byte_fetch_amount)
                data += curr_data_buffer

            request = RequestParser(data).parse()
            logger.debug(f"Finished building request object {request}")
            response = self.response_builder(request)
            logger.debug(f"Finished building response object {response}")
            response_data = response.data()
            self.send(response_data)
            self.close()



import inspect
import json
from abc import ABC

import jsonpickle

import web_framework_v2.http.http_request as http_request


class Annotation(ABC):
    def __init__(self, parameter_type, use_json_object_hook=False):
        """
        An annotation is what precedes the parameter name.
        method(self, parameter: annotation).
        this class is meant to be a base class for all custom annotations in the web_framework_v2.

        :param parameter_type: the type of the parameter. the data the annotation supplies will try to be converted to this type.
        """

        self._parameter_type = parameter_type
        self._use_json_object_hook = use_json_object_hook

    def value_generator(self, request: http_request.HttpRequest):
        raise NotImplementedError("Implement value_generator in non abstract class.")

    def adapt(self, data, parameter_type):
        try:
            if not self._use_json_object_hook:
                mapped_data = jsonpickle.decode(data)
                if type(mapped_data) is dict:
                    module = self._parameter_type.__module__
                    mapped_data[
                        "py/object"] = module + '.' + self._parameter_type.__qualname__ if module != "builtins" else self._parameter_type.__qualname__

                return jsonpickle.decode(json.dumps(mapped_data), classes=parameter_type)
            else:
                return json.loads(data, object_hook=lambda json_map: Annotation.json_object_hook(json_map, parameter_type))
        except Exception:
            return data

    @staticmethod
    def json_object_hook(json_map, parameter_type):
        if hasattr(parameter_type, "json_object_hook"):
            return parameter_type.json_object_hook(json_map)
        elif parameter_type is map:
            return json_map

        argspec = inspect.getfullargspec(parameter_type.__init__)
        argspec.args.pop(0)
        new_c = parameter_type(**{arg_name: json_map.get(arg_name, None) for arg_name in argspec.args})
        new_c.__dict__ = json_map
        return new_c


class QueryParameter(Annotation):
    def __init__(self, query_name: str, parameter_type=str, use_json_object_hook=False):
        self.query_name = query_name
        super().__init__(parameter_type, use_json_object_hook)

    def value_generator(self, request: http_request.HttpRequest):
        value = request.query_parameters.get(self.query_name, None) if request.query_parameters is not None else None

        if value is None:
            return None

        value = value if len(value) > 1 else value[0]

        return self.adapt(
            value,
            self._parameter_type
        )


class RequestBody(Annotation):
    def __init__(self, parameter_type=map, raw_format=False, use_json_object_hook=False):
        super().__init__(parameter_type, use_json_object_hook)
        self.raw_format = raw_format

    def value_generator(self, request):
        if self.raw_format:
            return request.body

        value = self.adapt(request.body, self._parameter_type) if request.body is not None else None
        return value if type(value) is not bytes else value.decode()


class PathVariable(Annotation):
    def __init__(self, variable_name, parameter_type=str, use_json_object_hook=False):
        super().__init__(parameter_type, use_json_object_hook)
        self.variable_name = variable_name

    def value_generator(self, request: http_request.HttpRequest):
        value = request.path_variables.get(self.variable_name, None)

        return self.adapt(value, self._parameter_type) if value is not None else None



import logging

from web_framework_v2.http import HttpRequest, ContentType
from web_framework_v2.http.http_method import HttpMethod
from web_framework_v2.http_server import HttpServer
from web_framework_v2.route import Endpoint
from web_framework_v2.route.endpoint_map import EndpointMap

logger = logging.getLogger(__name__)


class Framework:
    def __init__(self, static_folder: str, static_url_path: str, log_level=logging.INFO):
        logging.getLogger("web_framework_v2").setLevel(log_level)

        self._static_folder = static_folder
        self._static_url_path = static_url_path
        self._active = False
        self._http_server = HttpServer(self)
        self._endpoint_map = EndpointMap()

    def start(self):
        self._active = True
        self._http_server.start()

    def get_endpoint(self, request: HttpRequest):
        return self._endpoint_map.get_endpoint(request)

    def add_endpoint(self, route: str, func, methods: {HttpMethod}, match_headers: dict, content_type: ContentType):
        for method in methods:
            self._endpoint_map.add_route(Endpoint(route, method, content_type, func, match_headers))

    def endpoint(self, route: str, methods: {HttpMethod} = None, content_type: ContentType = ContentType.json, match_headers: dict = None):
        assert route is not None and type(route) is str, "Route must be a valid string!"

        if methods is None:
            methods = {HttpMethod.GET}

        if match_headers is None:
            match_headers = dict()

        if content_type is None:
            content_type = ContentType.json

        def decorator(f):
            self.add_endpoint(route, f, methods, match_headers, content_type)
            return f

        return decorator

    def get(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.GET}, content_type, match_headers)

    def post(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.POST}, content_type, match_headers)

    def put(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.PUT}, content_type, match_headers)

    def patch(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.PATCH}, content_type, match_headers)

    def delete(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.DELETE}, content_type, match_headers)

    def copy(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.COPY}, content_type, match_headers)

    def head(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.HEAD}, content_type, match_headers)

    def options(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.OPTIONS}, content_type, match_headers)

    def link(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.LINK}, content_type, match_headers)

    def unlink(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.UNLINK}, content_type, match_headers)

    def purge(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.PURGE}, content_type, match_headers)

    def lock(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.LOCK}, content_type, match_headers)

    def unlock(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.UNLOCK}, content_type, match_headers)

    def propfind(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.PROPFIND}, content_type, match_headers)

    def view(self, route: str, match_headers: dict = None, content_type: ContentType = ContentType.json):
        return self.endpoint(route, {HttpMethod.VIEW}, content_type, match_headers)

    def static_folder(self):
        return self._static_folder

    def static_url_path(self):
        return self._static_url_path

    def is_active(self):
        return self._active



import inspect
import logging

import jsonpickle

import web_framework_v2.annotations
import web_framework_v2.http.http_request as http_request
import web_framework_v2.decorator as decorator_module
from web_framework_v2.http import ContentType, http_response

logger = logging.getLogger(__name__)


class Method:
    encodable_content_types = [ContentType.json, ContentType.text]

    def __init__(self, method):
        """
        Wraps a web_framework_v2 function that has a route into a class that handles
        the execution of the function.
        :param method: the function
        """
        self._method = method

    def execute(self, request: http_request.HttpRequest, response):
        argspec = inspect.getfullargspec(self._method)
        args_len = len(argspec.args)
        defaults_len = len(argspec.defaults) if argspec.defaults is not None else 0
        kwargs = dict()
        defaults_map = dict()
        decorator_result_map = dict()

        # Build defaults map
        if defaults_len > 0:
            for i, default_value in enumerate(argspec.defaults):
                defaults_map[argspec.args[args_len - defaults_len + i]] = default_value

        # Execute decorator functionality
        if hasattr(self._method, "decorators"):
            decorators = getattr(self._method, "decorators", [])
            logger.debug(f"Method's decorators: {decorators}")
            request_body = web_framework_v2.annotations.RequestBody().value_generator(request)

            decorator: decorator_module.Decorator
            for decorator in decorators:
                should_exec, result, data = \
                    decorator.should_execute_endpoint(request, request_body)
                if not should_exec:
                    logger.debug(f"Decorator failed. Calling on_fail. {decorator}")
                    return decorator.on_fail(request, response, data)

                if result is not None:
                    decorator_result_map[type(decorator)] = result  # Used when building kwargs to set result based on annotation

        # Build kwargs
        if args_len > 0 and len(argspec.annotations) > 0:
            for parameter_name, annotation in argspec.annotations.items():
                if issubclass(type(annotation), web_framework_v2.annotations.Annotation):
                    kwargs[parameter_name] = annotation.value_generator(request)
                    if kwargs[parameter_name] is None and parameter_name in defaults_map:
                        kwargs[parameter_name] = defaults_map[parameter_name]
                else:
                    if annotation is http_request.HttpRequest or type(annotation) is http_request.HttpRequest:
                        kwargs[parameter_name] = request
                        defaults_map.pop(parameter_name, None)
                    elif annotation is http_response.HttpResponse or type(annotation) is http_response.HttpResponse:
                        kwargs[parameter_name] = response
                        defaults_map.pop(parameter_name, None)
                    elif annotation in decorator_result_map:
                        kwargs[parameter_name] = decorator_result_map[annotation]
                        defaults_map.pop(parameter_name, None)

        # Build defaults
        if defaults_len > 0:
            for parameter_name, default_value in defaults_map.items():
                if parameter_name in kwargs:
                    continue

                kwargs[parameter_name] = default_value

        logger.debug(f"Finished building method kwargs. {kwargs}")
        method_result = self._method(**kwargs)
        return jsonpickle.encode(method_result, unpicklable=False) if response.content_type in self.encodable_content_types else method_result



__all__ = ["ContentType", "HttpStatus", "HttpMethod", "HttpRequest", "HttpResponse"]

from .content_type import ContentType
from .http_status import HttpStatus
from .http_method import HttpMethod
from .http_request import HttpRequest
from .http_response import HttpResponse



from web_framework_v2.http import HttpMethod, HttpRequest


class RequestParser:
    def __init__(self, request: bytes):
        """
        Efficiently parses a request into a HttpRequest object.
        :param request: the HTTP request received from the browser
        """

        self.request = request
        self.length = len(self.request)
        self.cursor = 0

    def parse(self):
        try:
            method = HttpMethod.__getitem__(self.extract_till_space())
            url, query_parameters = self.parse_url()
            version = self.extract_till_carriage()
            headers = self.parse_header_space()
            self.skip_carriage()
            body = self.request[self.cursor:]

            return HttpRequest(method, url, version, headers, query_parameters, body)
        except:
            print("ERROR:")
            print(self.request)
            print("=-=-=-=-= ERROR =-=-=-=-=")
            raise

    def current(self):
        if self.cursor >= self.length:
            return None
        return chr(self.request[self.cursor])

    def next(self):
        if self.cursor + 1 >= self.length:
            return None
        return chr(self.request[self.cursor + 1])

    def parse_url(self):
        url_buffer = ""
        query_parameters = {}

        while self.current() != "?" and self.current() != " ":
            url_buffer += self.current()
            self.cursor += 1

        if self.current() != "?":
            return url_buffer, None

        self.cursor += 1
        curr_name = ""
        searching_name = True
        while self.current() != " ":
            curr = self.current()
            if curr == "&":
                searching_name = True
                if curr_name not in query_parameters and curr_name != "":
                    query_parameters[curr_name] = None
                curr_name = ""
                self.cursor += 1
                continue
            elif curr == "=":
                searching_name = False
                self.cursor += 1
                continue

            if searching_name:
                curr_name += curr
                self.cursor += 1
            else:
                values_str = ""
                while self.current() != "&" and self.current() != " ":
                    values_str += self.current()
                    self.cursor += 1
                curr_value = values_str.split(",")
                query_parameters[curr_name] = curr_value

        if curr_name not in query_parameters:
            query_parameters[curr_name] = None
        self.cursor += 1
        return url_buffer, query_parameters

    def parse_header_space(self):
        headers = {}

        while self.skip_carriage() < 2:
            header = self.extract_till_carriage()
            name = ""
            idx = 0

            while header[idx] != ":":
                name += header[idx]
                idx += 1
            idx += 1
            while header[idx] == " ":
                idx += 1
            value = header[idx:]
            headers[name] = value

        return headers

    def extract_till_space(self):
        buffer = ""
        while self.current() != " ":
            buffer += self.current()
            self.cursor += 1

        self.cursor += 1
        return buffer

    def extract_till_carriage(self):
        buffer = ""

        while self.cursor + 4 <= self.length:
            if self.current() == '\r' \
                    and self.next() == '\n':
                return buffer

            buffer += self.current()
            self.cursor += 1

        return buffer

    def skip_carriage(self):
        match_count = 0

        while self.current() == "\r" and self.next() == "\n":
            self.cursor += 2
            match_count += 1

        return match_count



__all__ = ["ContentType", "HttpStatus", "HttpMethod", "HttpRequest", "HttpResponse",
           "Annotation", "QueryParameter", "RequestBody", "PathVariable", "Decorator", "JwtSecurity",
           "JwtTokenFactory", "JwtTokenAuth", "HttpClient", "HttpServer", "Framework"]

from .http import *
from .decorator import Decorator
from .annotations import Annotation, QueryParameter, RequestBody, PathVariable
from .jwt import JwtSecurity, JwtTokenFactory, JwtTokenAuth
from .http_client import HttpClient
from .http_server import HttpServer
from .framework import Framework



from enum import Enum


class ContentType(Enum):
    html = "text/html; charset=utf-8"
    htm = "text/html; charset=utf-8"
    txt = "text/html; charset=utf-8"
    text = "text/html; charset=utf-8"
    js = "text/javascript; charset=utf-8"
    css = "text/css"
    jpg = "image/jpeg"
    ico = "image/vnd.microsoft.icon"
    gif = "image/gif"
    json = "application/json"
    xml = "application/xml"

    def __str__(self):
        return "content-type: " + self.value



import re

import web_framework_v2.http.http_request as http_request
from web_framework_v2.http import HttpMethod, ContentType
import web_framework_v2.method as method_module


class Endpoint:
    VARIABLE_MATCHER = re.compile(r"({.+?})")
    SLASH_EXTRACTOR = re.compile(r"/?([^/]+)/?")

    def __init__(self, route: str, http_method: HttpMethod, content_type: ContentType, func, match_headers: dict):
        if len(route) == 0:
            route = "/"

        route = route if route[0] == "/" else "/" + route
        route = route if route[-1] == "/" else route + "/"

        self._route = route
        self._http_method = http_method
        self._content_type = content_type
        self._func = func
        self._match_headers = match_headers
        self._method = method_module.Method(self._func)

        self._variable_table = {i.group(): i.span() for i in self.VARIABLE_MATCHER.finditer(self._route)}
        self._route_contains_variables = len(self._variable_table) > 0
        self._route_slashes = self.SLASH_EXTRACTOR.findall(self._route)

    def execute(self, request: http_request.HttpRequest, response, path_variables):
        request.path_variables = path_variables
        return self._method.execute(request, response)

    def matches_headers(self, request: http_request.HttpRequest):
        if self._match_headers is None or len(self._match_headers) == 0:
            return True

        request_headers = request.headers
        matches = True

        for header, value in self._match_headers.items():
            matches = header in request_headers and request_headers[header] == value

            if not matches:
                break

        return matches

    def matches_url(self, url):
        if not self.has_route_variables():
            url = url if url[-1] == "/" else url + "/"
            url = url if url[0] == "/" else "/" + url
            return url == self._route, None

        variable_values = {}
        url_slashes = self.SLASH_EXTRACTOR.findall(url)

        if len(url_slashes) != len(self._route_slashes):
            return False, None

        for url_slash, route_slash in zip(url_slashes, self._route_slashes):
            if route_slash[0] != "{" and route_slash[-1] != "}":
                if url_slash != route_slash:
                    return False, None
            else:
                variable_values[route_slash[1:-1]] = url_slash

        return True, variable_values

    def has_route_variables(self):
        return self._route_contains_variables

    def route(self):
        return self._route

    def method(self):
        return self._http_method

    def content_type(self):
        return self._content_type

    def __str__(self):
        return f"Route(url: {self.route()})"



from __future__ import annotations

from . import Endpoint
import web_framework_v2.http.http_request as http_request


class EndpointMap:
    def __init__(self):
        self._method_routes_map = {}  # HttpMethod: {route_str: Route}

    def get_endpoint(self, request: http_request.HttpRequest) -> tuple[Endpoint | None, str | None] | None:
        endpoint_obj = self._method_routes_map.get(request.method, {}).get(request.url, None)

        if endpoint_obj is None:
            method_routes = self._method_routes_map.get(request.method, None)

            if method_routes is None:
                return None, None

            for route in method_routes.values():
                matches, variable_map = route.matches_url(request.url)
                if matches and route.matches_headers(request):
                    return route, variable_map

        if endpoint_obj is not None and not endpoint_obj.matches_headers(request):
            return None, None

        return endpoint_obj, None

    def add_route(self, route: Endpoint):
        assert self._method_routes_map.get(route.method(), {}).get(route.route(), None) is None, \
            "Route already exists! Cannot add existing route!"

        self._method_routes_map.setdefault(route.method(), {})
        self._method_routes_map[route.method()][route.route()] = route

    def __str__(self):
        return str(self._method_routes_map)



from abc import ABC
from typing import Tuple


class Decorator(ABC):
    """
    Decorate a route to change it's pre-execution functionality
    """

    def should_execute_endpoint(self, request, request_body) -> Tuple[bool, object, object]:
        raise NotImplementedError(f"Must implement should_execute_route() in {type(self)}")

    def on_fail(self, request, response, data):
        pass

    def __call__(self, fun):
        decorators = getattr(fun, "decorators", [])
        decorators.append(self)
        setattr(fun, "decorators", decorators)

        return fun



import logging
import socket
import threading

from web_framework_v2.http import HttpResponse
from web_framework_v2.http_client import HttpClient

logger = logging.getLogger(__name__)


class HttpServer:
    def __init__(self, framework):
        self._framework = framework
        self._ip = ("0.0.0.0", 80)
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._client_listen_thread = threading.Thread(target=self.__client_listener)

    def start(self):
        self._socket.bind(self._ip)
        self._socket.listen(2)
        self._client_listen_thread.start()
        logger.info("The server is active and listening...")

    def __client_listener(self):
        while self._framework.is_active():
            client_socket, address = self._socket.accept()
            logger.debug(f"Accepted client connection from {address}")
            client = HttpClient(client_socket, address, lambda req: self.response_builder(req))
            client.start()

        self.shutdown()

    def response_builder(self, request):
        logger.debug(f"Attempting to find a endpoint using {request.url}")
        route, path_variables = self._framework.get_endpoint(request)

        if route is not None:
            logger.debug(f"Found route {route}")
            return HttpResponse.build_from_route(request, route, path_variables)
        else:
            logger.debug("Failed to find endpoint, building response using static folder.")
            path = self._framework.static_folder() + (self._framework.static_url_path() if request.url == "/" else request.url)
            return HttpResponse.build_from_file(request, path)

    def shutdown(self):
        self._socket.shutdown(socket.SHUT_WR)



import unittest

from web_framework_v2.http import HttpMethod
from web_framework_v2.route import Endpoint


class RouteMatching(unittest.TestCase):
    mock_routes = [
        "hockey/database/report",
        "/hockey/database/report",
        "hockey/database/report/",
        "/hockey/database/report/",

        "hockey/player/{name}/score",
        "/hockey/player/{name}/score",
        "hockey/player/{name}/score/",
        "/hockey/player/{name}/score/",

        "hockey/player/{name}",
        "/hockey/player/{name}",
        "hockey/player/{name}/",
        "/hockey/player/{name}/",

        "{game}/player/{name}",
        "/{game}/player/{name}",
        "{game}/player/{name}/",
        "/{game}/player/{name}/",

        "{game}/player/{name}/score",
        "/{game}/player/{name}/score",
        "{game}/player/{name}/score/",
        "/{game}/player/{name}/score/",
    ]

    corresponding_test_routes = [
        "hockey/database/report",
        "/hockey/database/report",
        "hockey/database/report/",
        "/hockey/database/report/",

        "hockey/player/HeKNon/score",
        "/hockey/player/HeKNon/score",
        "hockey/player/HeKNon/score/",
        "/hockey/player/HeKNon/score/",

        "hockey/player/HeKNon",
        "/hockey/player/HeKNon",
        "hockey/player/HeKNon/",
        "/hockey/player/HeKNon/",

        "softBaLl/player/HeKNon",
        "/softBaLl/player/HeKNon",
        "softBaLl/player/HeKNon/",
        "/softBaLl/player/HeKNon/",

        "softBaLl/player/HeKNon/score",
        "/softBaLl/player/HeKNon/score",
        "softBaLl/player/HeKNon/score/",
        "/softBaLl/player/HeKNon/score/",
    ]
    routes = [Endpoint(route, HttpMethod.GET, None) for route in mock_routes]

    def test_route_matching(self):
        for (i, route) in enumerate(self.routes):
            self.assertEqual(
                route.matches_url(self.corresponding_test_routes[i])[0],
                True,
                '"' + self.corresponding_test_routes[i] + '" doesn\'t match route "' + self.mock_routes[i] + '"'
            )

    def test_path_variable_matching(self):
        for (i, route) in enumerate(self.routes):
            if i < 4:
                self.assertEqual(
                    route.matches_url(self.corresponding_test_routes[i])[1],
                    None
                )

                continue
            self.assertEqual(
                route.matches_url(self.corresponding_test_routes[i])[1].get("name"),
                "HeKNon"
            )

            if i > 11:
                self.assertEqual(
                    route.matches_url(self.corresponding_test_routes[i])[1].get("game"),
                    "softBaLl"
                )


if __name__ == '__main__':
    unittest.main()
